import tensorflow as tf
from ..base import DatasetTFRecords, NodeKeys


class MNISTTFRecords(DatasetTFRecords):
    def __init__(self, name='dataset', *, files=None, batch_size=None, one_hot=None, **config):
        super(__class__, self).__init__(name, files=files,
                                        batch_size=batch_size,
                                        one_hot=one_hot,
                                        **config)

    @classmethod
    def _default_config(cls):
        return {
            'files': ['/home/hongxwing/Datas/mnist/mnist.train.tfrecord'],
            'batch_size': 32,
            'one_hot': True,
        }

    @classmethod
    def _parse_example(cls, example):
        return tf.parse_single_example(example, features={
            'image': tf.FixedLenFeature([], tf.string),
            'shape': tf.FixedLenFeature([2], tf.int64),
            'label': tf.FixedLenFeature([1], tf.int64)
        })

    @classmethod
    def _decode_image(cls, data):
        return {'image': tf.decode_raw(data['image'], tf.uint8),
                'shape': data['shape'],
                'label': data['label']}

    @classmethod
    def _change_dtype(cls, data):
        return {'image': tf.to_float(data['image']),
                'shape': data['shape'],
                'label': data['label']}

    @classmethod
    def _reshape_tensors(cls, data):
        return {'image': tf.reshape(data['image'], [28, 28, 1]),
                'shape': tf.concat([data['shape'], [1]], axis=0),
                'label': tf.reshape(data['label'], [])}

    def _normalize(self, data):
        if self._normalizer is None:
            image = data['image']
        else:
            image = self._normalizer(data['image'])[NodeKeys.MAIN]
        return {'image': image,
                'shape': data['shape'],
                'label': data['label']}
        return data

    def _maybe_onehot(self, data):
        result = {
            'image': data['image'],
            'shape': data['shape']
        }
        if self.c['one_hot']:
            result.update({'label': tf.one_hot(data['label'], depth=10)})
        else:
            result.update({'label': data['label']})
        return result

    def _processing(self):
        dataset = super()._processing()
        return (dataset
                .map(self._parse_example)
                .map(self._decode_image)
                .map(self._change_dtype)
                .map(self._reshape_tensors)
                # .map(self._normalize)
                .map(self._maybe_onehot)
                .batch(self.c['batch_size'])
                .cache()
                .repeat())


# from ..base import Graph


# class MNISTLoadAll(Graph):
#     def __init__(self, name='/dataset', **config):
#         # from ..preprocessing.normalizer import SelfMinMax

#         super(__class__, self).__init__(name, **config)
#         self._normalizer = None
#         method = self.c['normalization']['method'].lower()
#         if method != 'pass':
#             self._normalizer = get_normalizer(
#                 method)(self.name / 'normalization', )
#         self.register_main_node(self.__load_data())
#         self.register_node('image', self.as_tensor()['image'])
#         self.register_node('label', self.as_tensor()['label'])

#     def get_feed_dict(self):
#         return tf.get_default_session().run(
#             {'image': self.tensor('image'), 'label': self.tensor('label')})

#     @classmethod
#     def _default_config(cls):
#         return {
#             'batch_size': 32,
#         }

#     def post_session_created(self):
#         tf.get_default_session().run(self._iter_init,
#                                      feed_dict={
#                                          self.tensor('images'): self.images,
#                                          self.tensor('labels'): self.labels
#                                      })

#     def _normalize(self, image):
#         if self._normalizer is None:
#             return image
#         else:
#             return self._normalizer(image)['data']

#     def __load_data(self):
#         from tensorflow.examples.tutorials.mnist import input_data
#         from tensorflow.contrib.data import Dataset
#         mnist = input_data.read_data_sets(
#             '/home/hongxwing/Datas/mnist/tfdefault', one_hot=True)
#         self.images = mnist.train.images
#         self.labels = mnist.train.labels

#         # def gen_image():
#         #     for i in range(images.shape[0]):
#         #         yield {'image': images[i, ...], 'label': label[i, ...]}
#         with tf.name_scope(self.basename):
#             images_tensor = tf.placeholder(tf.float32, self.images.shape,
#                                            'mnist_images')
#             self.register_node('images', images_tensor)
#             labels_tensor = tf.placeholder(tf.int64, self.labels.shape,
#                                            'mnist_label')
#             self.register_node('labels', labels_tensor)
#             dataset_image = (Dataset.from_tensor_slices(images_tensor)
#                              .map(lambda img: tf.reshape(img, [28, 28, 1]))
#                              .map(self._normalize))
#             dataset_label = Dataset.from_tensor_slices(labels_tensor)
#             dataset = Dataset.zip(
#                 {'image': dataset_image, 'label': dataset_label})
#             dataset = dataset.batch(self.param('batch_size')).repeat()
#             iterator = dataset.make_initializable_iterator()
#             self._iter_init = iterator.initializer
#             next_element = iterator.get_next()
#         return next_element



from dxpy.filesystem import Path
import os
datasets_configs = {
    
    'dataset_root_path': os.environ.get('PATHS_DATASET', str(Path(os.environ.get('HOME')) / 'Datas')),
    'analytical_phantom_sinogram': {
        'path': '/home/hongxwing/Datas/Phantom',
    },
    'apssr': {
        'image_type': 'sinogram',
        'target_shape': [320, 320],
        'super_resolution': {
            'nb_down_sample': 3
        },
    }
}
config = {
    'train': {
        'summary_freq': 60,
        'ckpt_path': './save'
    },
    'datasets': datasets_configs
}


def get_config():
    from dxpy.configs import ConfigsView
    return ConfigsView(config)

def get_configs_view():
    from dxpy.configs import ConfigsView
    return ConfigsView(config)

def clear_config():
    global config
    for k in config:
        config.pop(k)



import numpy as np
import tensorflow as tf
import numpy as np


def ensure_shape(input_tensor, shape=None, *, batch_size=None):
    """
    Reshpe input_tensor to given shape.
    """
    if isinstance(input_tensor, np.ndarray):
        return np.copy(input_tensor)
    with tf.name_scope('shape_ensure'):
        shape_input = input_tensor.shape.as_list()
        if shape is None:
            shape = shape_input
        if shape[0] is None:
            shape[0] = batch_size
        return tf.reshape(input_tensor, shape)


import typing


def to_tensor(input_: typing.TypeVar('TensorConvertable', tf.Tensor, np.ndarray),
              *,
              name: str="to_tensor",
              restrict_tensorflow=False,
              default_device='cpu') -> tf.Tensor:
    """
    Helper function to convert multiple kind of Tensor-convertable objects to Tensor.
    """
    if isinstance(input_, Graph):
        return input_.as_tensor()
    if isinstance(input_, np.ndarray) and not restrict_tensorflow:
        return input_
    if isinstance(input_, tf.Tensor):
        return input_
    # with tf.name_scope()


def to_tensor_with_shape(input_, shape: typing.Tuple[int], *, batch_size: int=None, name: str='to_tensor_with_shape') -> tf.Tensor:
    if shape is None and batch_size is not None:
        shape = shape_as_list(input_)
        shape[0] = batch_size

    input_ = to_tensor(input_)
    if isinstance(input_, np.ndarray):
        return input_.reshape(shape)
    elif isinstance(input_, tf.Tensor):
        with tf.name_scope(name):
            return tf.reshape(input_, shape)


def shape_as_list(input_) -> typing.List[int]:
    if isinstance(input_, tf.Tensor):
        return list(input_.shape.as_list())
    if isinstance(input_, np.ndarray):
        return list(input_.shape)
    if isinstance(input_, (tuple, list)):
        return list(input_)




"""
Super resolution dataset for analytical phantom sinogram dataset
"""
from typing import List

import tensorflow as tf

from dxpy.configs import configurable

from ...config import config
from ...graph import Graph, NodeKeys
from ..raw.analytical_phantom_sinogram import Dataset


class AnalyticalPhantomSinogramDatasetForSuperResolution(Graph):
    @configurable(config, with_name=True)
    def __init__(self, name='datasets/apssr',
                 image_type: str='sinogram',
                 nb_down_sample: int=3,
                 *,
                 log_scale: bool=False,
                 with_white_normalization: bool=True,
                 with_poission_noise: bool=False,
                 with_noise_label: bool=False,
                 target_shape: List[int]=None,
                 **kw):
        """
        Args:
            -   image_type: 'sinogram' or 'image'
            -   log_scale: produce log datas
            -   with_white_normalization: normalize by mean and std
            -   with_poission_noise (apply to sinogram only): perform poission sample
            -   target_shape: (apply to sinogram only), random crop shape
        Returns:
            a `Graph` object, which has several nodes:
        Raises:
        """
        super().__init__(name, image_type=image_type,
                         log_scale=log_scale,
                         with_poission_noise=with_poission_noise,
                         target_shape=target_shape,
                         with_white_normalization=with_white_normalization,
                         with_noise_label=with_noise_label,
                         nb_down_sample=nb_down_sample, **kw)
        from ...model.normalizer.normalizer import FixWhite, ReduceSum
        from ...model.tensor import ShapeEnsurer
        from dxpy.core.path import Path
        name = Path(name)
        if image_type == 'sinogram':
            fields = ['sinogram']
        else:
            fields = ['phantom'] + ['recon{}x'.format(2**i)
                                    for i in range(self.param('nb_down_sample') + 1)]
        with tf.name_scope('aps_{img_type}_dataset'.format(img_type=image_type)):
            dataset_origin = Dataset(
                self.name / 'analytical_phantom_sinogram', fields=fields)
            if image_type == 'sinogram':
                dataset = self._process_sinogram(dataset_origin)
            else:
                dataset = self._process_recons(dataset_origin)
        # for k in dataset:
            # self.register_node(k, dataset[k])
        for i in range(self.param('nb_down_sample') + 1):
            self.register_node('input/image{}x'.format(2**i), dataset['input/image{}x'.format(2**i)])
            if self.param('image_type') == 'image' and i == 0:
                self.register_node('label/image1x', dataset['label/phantom'])
                continue
            if self.param('with_noise_label'):
                self.register_node('label/image{}x'.format(2**i), dataset['input/image{}x'.format(2**i)])
            else:
                self.register_node('label/image{}x'.format(2**i), dataset['label/image{}x'.format(2**i)])
   

    def _process_sinogram(self, dataset):
        from ...model.normalizer.normalizer import ReduceSum, FixWhite
        from ..super_resolution import SuperResolutionDataset
        from ...utils.tensor import shape_as_list
        if self.param('log_scale'):
            stat = dataset.LOG_SINO_STAT
        else:
            stat = dataset.SINO_STAT
        # dataset = ReduceSum(self.name / 'reduce_sum', dataset['sinogram'],
            # fixed_summation_value=1e6).as_tensor()
        dataset = dataset['sinogram']
        if self.param('with_poission_noise'):
            with tf.name_scope('add_with_poission_noise'):
                noise = tf.random_poisson(dataset, shape=[])
                dataset = tf.concat([noise, dataset], axis=0)
        if self.param('log_scale'):
            dataset = tf.log(dataset + 0.4)
        if self.param('with_white_normalization'):
            dataset = FixWhite(name=self.name / 'fix_white',
                               inputs=dataset, mean=stat['mean'], std=stat['std']).as_tensor()
        dataset = tf.random_crop(dataset,
                                 [shape_as_list(dataset)[0]] + list(self.param('target_shape')) + [1])
        dataset = SuperResolutionDataset(self.name / 'super_resolution',
                                         lambda: {'image': dataset},
                                         input_key='image',
                                         nb_down_sample=self.param('nb_down_sample'))
        keys = ['image{}x'.format(2**i)
                for i in range(dataset.param('nb_down_sample') + 1)]
        if self.param('with_poission_noise'):
            result = {
                'input/' + k: dataset[k][:shape_as_list(dataset[k])[0] // 2, ...] for k in keys}
            result.update(
                {'label/' + k: dataset[k][shape_as_list(dataset[k])[0] // 2:, ...] for k in keys})
        else:
            result = {'input/' + k: dataset[k] for k in keys}
            result.update({'label/' + k: dataset[k] for k in keys})
        return result

    def _process_recons(self, dataset):
        from ...model.normalizer.normalizer import ReduceSum, FixWhite
        from ..super_resolution import SuperResolutionDataset
        from ...utils.tensor import shape_as_list
        keys = ['recon{}x'.format(2**i)
                for i in range(self.param('nb_down_sample') + 1)]
        if self.param('log_scale'):
            stat = dataset.LOG_RECON_STAT
        else:
            stat = dataset.RECON_STAT
        phantom = dataset['phantom']
        dataset = {k: dataset[k] for k in keys}
        for i, k in enumerate(keys):
            if i == 0:
                continue
            dataset[k] = tf.nn.avg_pool(dataset[k],
                                        [1] + [2**i, 2**i] + [1],
                                        padding='SAME',
                                        strides=[1] + [2**i, 2**i] + [1]) * (2.0**i * 2.0**i)

        # dataset = {k: ReduceSum(self.name / 'reduce_sum' / k, dataset[k], fixed_summation_value=1e6).as_tensor() for k in keys}
        if self.param('log_scale'):
            for k in keys:
                dataset[k] = tf.log(dataset[k] + 1.0)
            phantom = tf.log(phantom + 1.0)
        if self.param('with_white_normalization'):
            for i, k in enumerate(keys):
                dataset[k] = FixWhite(name=self.name / 'fix_white' / k,
                                      inputs=dataset[k], mean=stat['mean'], std=stat['std']).as_tensor()
            phantom = FixWhite(name=self.name / 'fix_white' / 'phantom',
                               inputs=dataset[k], mean=stat['mean'], std=stat['std']).as_tensor()

        result = dict()
        for i, k in enumerate(keys):
            result['input/image{}x'.format(2**i)] = dataset[k]
            result['label/image{}x'.format(2**i)
                   ] = result['input/image{}x'.format(2**i)]
        result['label/phantom'] = phantom
        return result



import tensorflow as tf

from ..graph import Graph
from ..scalar import ScalarVariable
from ..model.tensor import shape_as_list

from dxpy.configs import configurable
from dxpy.learn.config import config

import typing
from dxpy.learn.session import get_default_session

class Trainer(Graph):
    """
    Trainer which make it easier to:
        1. Unified access for single GPU and multi GPU (auto detect);
        2. Changing learning rate made easy;
        3. Training partial of variables easily (TODO)
    """

    def _check_inputs(self, loss, variables):
        def _check_tensor(l):
            if isinstance(l, tf.Tensor):
                shape = shape_as_list(l)
                if len(shape) > 1 or (len(shape) == 1 and shape[0] > 1):
                    msg = "Loss of trainer can only be scalar or list of scalar, got loss (or element of loss) of shape {}."
                    raise ValueError(msg.format(shape))
            else:
                raise TypeError(
                    "Loss (or element of loss) of trainer can only be scalar or list of scalar, not {}.".format(type(loss)))
        if loss is None:
            raise TypeError(
                "Loss of trainer can only be scalar or list of scalar, not None.")
        if isinstance(loss, (list, tuple)):
            for l in loss:
                _check_tensor(l)
        else:
            _check_tensor(loss)
    def _norm_learning_rate(self, base_rate):
        return base_rate * self.param('batch_size') 
    @configurable(config, with_name=True)
    def __init__(self, name: str=None, loss=None, variables=None, learning_rate=1e-3, optimizer='sgd', batch_size=1, **kw):
        """
        Inputs:
            loss: scalar or list of scalar (multi gpu)
            variables: list of tensors
        """
        super(__class__, self).__init__(
            name, learning_rate=learning_rate, optimizer=optimizer, batch_size=batch_size, **kw)
        self._check_inputs(loss, variables)
        self.loss = loss
        if variables is None:
            variables = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES)
        self.variables = variables
        with tf.variable_scope(self.basename):
            self.register_node('learning_rate', ScalarVariable(self.name / 'learning_rate',
                                                               init_value=self.c['learning_rate']))
            self.optimizer = self._get_optimizer()
            self.register_main_node(self._get_train_step())
            self.register_main_task(self._train)
            self.register_task('multiply_learning_rate',
                               self._multiply_learning_rate)
            self.register_task('set_learning_rate', self._set_learning_rate)

    def _train(self, feeds):
        sess = get_default_session()
        sess.run(self.as_tensor(), feed_dict=feeds)

    def _multiply_learning_rate(self, feeds):
        old_value = self.nodes['learning_rate'].get_value()
        new_value = old_value * self.param('ratio', feeds)
        self.nodes['learning_rate'].set_value(new_value)

    def decay_learning_rate(self, ratio):
        self.run('multiply_learning_rate', feeds={'ratio': ratio})

    def _set_learning_rate(self, feeds):
        self.nodes['learning_rate'].set_value(self.param('learning_rate'))

    def _get_optimizer(self):
        if self.param('optimizer').lower() == 'sgd':
            return tf.train.GradientDescentOptimizer(self.nodes['learning_rate'].as_tensor())
        elif self.param('optimizer') == 'adam':
            return tf.train.AdamOptimizer(self.nodes['learning_rate'].as_tensor())
        else:
            raise ValueError('Unknown optimizer type {}.'.format(
                self.param('optimizer')))

    # def _get_gradients(self):
    #     from ..utils.general import device_name
    #     results = []
    #     if isinstance(self.loss, (list, tuple)):
    #         for i in range(self.c['nb_gpu']):
    #             with tf.device(device_name('gpu', i)):
    #                 results.append(self.optimizer.compute_gradients(
    #                     self.loss[i], self.variables))
    #     else:
    #         results.append(self.optimizer.compute_gradients(
    #             self.loss, self.variables))
    #     return results

    # @classmethod
    # def __analysis_grad_and_vars(cls, grad_and_vars):
    #     nb_gpu = len(grad_and_vars)
    #     variables = grad_and_vars[0][1]
    #     grads = [g for g, _ in grad_and_vars if g is not None]
    #     if len(grads) > 0:
    #         grad = tf.add_n(grads) / nb_gpu
    #     else:
    #         grad = None
    #     return nb_gpu, grad, variables

    # def __maybe_clip(self, grad):
    #     if self.c.get('with_clip', False):
    #         grad = tf.clip_by_value(grad,
    #                                 -self.c['clipv'],
    #                                 self.c['clipv'])
    #     return grad

    # def _add_summary(self, average_grads):
    #     pass

    # def _get_train_step_full(self):
    #     from ..scalar import global_step
    #     tower_gradients = self._get_gradients()
    #     with tf.name_scope('train_step'), tf.device('/cpu:0'):
    #         average_grads = []
    #         for grad_and_vars in zip(*tower_gradients):
    #             _, grad, var = self.__analysis_grad_and_vars(grad_and_vars)
    #             grad = self.__maybe_clip(grad)
    #             average_grads.append((grad, var))
    #         train_op = self.optimizer.apply_gradients(
    #             average_grads, global_step=global_step())
    #         self._add_summary(average_grads)
    #         return train_op

    def _get_train_step(self):
        from ..scalar import global_step
        return self.optimizer.minimize(self.loss, global_step(), self.variables)

    # def _get_train_step(self):
    #     if not isinstance(self.loss, (list, tuple)):
    #         return self._get_train_step_simple()
    #     else:
    #         return self._get_train_step_full()



import matplotlib.pyplot as plt
import numpy as np
from .. import image as nei


def grid_view(image_lists, windows=None, nb_column=8, scale=1.0, cmap=None, *, hide_axis=True, tight_c=0.01, return_figure=False, dpi=None, adjust_figure_size=True):
    """ subplot list of images of multiple categories into grid subplots
    Args:
        image_lists: list of [list of images or 4D tensor]
        windows: list of windows
        nb_column: columns of images
    Returns:
        Return figure if return_figure is true, else None.
    """
    if not isinstance(image_lists, (list, tuple)):
        raise TypeError(
            "image_lists should be list or tuple, got {}.".format(type(image_lists)))
    nb_cata = len(image_lists)
    if windows is None:
        windows = [(None, None)] * nb_cata
    image_lists = list(map(nei.try_unbatch, image_lists))
    for i, v in enumerate(image_lists):
        if not isinstance(v, list) and isinstance(v, np.ndarray) and v.ndim == 2:
            image_lists[i] = [v]
    image_lists = [list(map(nei.fix_dim, imgs)) for imgs in image_lists]

    nb_images = max([len(imgs) for imgs in image_lists])

    nb_row = np.ceil(nb_images / nb_column) * nb_cata

    def adjust_figure_size():
        width = nb_column * scale
        height = nb_row * scale
        if adjust_figure_size:
            shapes = [(il[0].shape[1], il[0].shape[0])
                      for il in image_lists if len(il) > 0]
            w, h = zip(*shapes)
            width *= np.mean(w) / 100
            height *= np.mean(h) / 100
        return (width, height)

    if dpi is None:
        dpi = 2000.0 / adjust_figure_size()[0]
    fig = plt.figure(figsize=adjust_figure_size(), dpi=dpi)

    for k in range(nb_cata):
        for i in range(nb_images):
            if i > len(image_lists[k]):
                continue
            r = i // nb_column * nb_cata + k
            c = i % nb_column
            ax = plt.subplot(nb_row, nb_column, r * nb_column + c + 1)
            plt.imshow(image_lists[k][i], cmap=cmap,
                       vmin=windows[k][0], vmax=windows[k][1])
            if hide_axis:
                ax.get_xaxis().set_visible(False)
                ax.get_yaxis().set_visible(False)
    if isinstance(tight_c, (list, tuple)):
        h_pad, w_pad = tight_c
    else:
        h_pad = tight_c
        w_pad = tight_c
    plt.tight_layout(h_pad=h_pad, w_pad=w_pad)
    if return_figure:
        return fig



from dxpy.file_system.path import Path
from dxpy.task.representation import creators


class SleepChainCreator:
    def __init__(self, workdir=None, duration=5):
        self.workdir = Path(workdir)
        self.duration = duration

    def create(self):
        cmd = 'sleep {0} && hostname'.format(self.duration)
        t1 = creators.task_command(
            cmd, workdir=self.workdir, desc='sleep chain #0')
        t2 = creators.task_command(
            cmd, workdir=self.workdir, desc='sleep chain #1')
        t3 = creators.task_command(
            cmd, workdir=self.workdir, desc='sleep chain #2')
        g = creators.task_graph([t1, t2, t3], [None, 0, 1])
        return g


class SbatchCreator:
    def __init__(self, workdir=None, ):
        pass



from pprint import pprint

import numpy as np
import tensorflow as tf
from dxpy.learn.session import Session
from dxpy.learn.train.summary import SummaryWriter
from dxpy.learn.utils.general import pre_work
from tqdm import tqdm
from dxpy.learn.config import config
import yaml
from dxpy.configs import configurable, ConfigsView


@configurable(ConfigsView(config).get('train'))
def get_train_configs(summary_freq=1000, save_freq=10000, steps=100000000):
    return {'summary_freq': summary_freq,
            'save_freq': save_freq,
            'steps': steps}


def train(definition_func):
    with open('dxln.yml') as fin:
        ycfg = yaml.load(fin)
    config.update(ycfg)
    pre_work()
    train_cfgs = get_train_configs()
    steps = train_cfgs['steps']
    summary_freq = train_cfgs['summary_freq']
    save_freq = train_cfgs['save_freq']
    network, summary = definition_func(ycfg)
    session = Session()
    with session.as_default():
        network.post_session_created()
        summary.post_session_created()
        session.post_session_created()

    with session.as_default():
        network.load()
        for i in tqdm(range(steps)):
            network.train()
            if i % summary_freq == 0 and i > 0:
                summary.summary()
                summary.flush()
            if i % save_freq == 0 and i > 0:
                network.save()

    with session.as_default():
        network.save()
        

def train_dist(definition_func):
    from dxpy.learn.utils.general import load_yaml_config
    load_yaml_config('dxln.yml')
    
